/**
  ******************************************************************************
  * @file         flash.c
  * @brief        This file provides code for the flash memory.
  *               
  ******************************************************************************
  * @attention
  *
  *
  ******************************************************************************
  */

/* Includes ------------------------------------------------------------------*/
#include <zephyr.h>
#include <drivers/spi.h>
#include "flash.h"
#include <string.h>

/* Private typedef -----------------------------------------------------------*/


/* Private define ------------------------------------------------------------*/

#define FLASH_TEST_DATA_SIZE 1024
/* Flags of SPI Transfer */
#define SPI_XFER_BEGIN		0x01	/* Assert CS before transfer */
#define SPI_XFER_END		0x02	/* Deassert CS after transfer */

/* Private macro -------------------------------------------------------------*/
#define le16_to_cpu(x) x
#define le32_to_cpu(x) x

/* Private function prototypes -----------------------------------------------*/
static void sanitize_string(char *s, size_t len);
static uint16_t onfi_crc16(uint16_t crc, uint8_t const *p, size_t len);

/* Private variables ---------------------------------------------------------*/
static const struct device * spi_dev;

static struct spi_cs_control spi_cs = {
    .gpio_dev = NULL,
    .gpio_pin = 18,
    .gpio_dt_flags = GPIO_ACTIVE_LOW,
    .delay = 0,
  };

static struct spi_config spi_cfg = {
	.operation = SPI_WORD_SET(8) | SPI_TRANSFER_MSB,// |
		    // SPI_MODE_CPOL | SPI_MODE_CPHA,
	.frequency = 8000000,
	.slave = 0,
	.cs = &spi_cs,
};

static const struct device *gpioPort0;

static uint8_t m_buffer_tx[FLASH_TEST_DATA_SIZE] = {0};
static uint8_t m_buffer_rx[FLASH_TEST_DATA_SIZE] = {0};

uint8_t buffer[256 * 3] = {0};

static bool flashSuccessfullyInitialized = false;

	static struct spi_buf tx_buf = {
		.buf = m_buffer_tx,
		.len = sizeof(m_buffer_tx)
	};
	static const struct spi_buf_set tx = {
		.buffers = &tx_buf,
		.count = 1
	};

	static struct spi_buf rx_buf = {
		.buf = m_buffer_rx,
		.len = sizeof(m_buffer_rx),
	};
	static const struct spi_buf_set rx = {
		.buffers = &rx_buf,
		.count = 1
	};

struct spi_nand_chip chip;

struct spi_nand_cmd_cfg {
	uint8_t		opcode;
	uint8_t		addr_bytes;
	uint8_t		addr_bits;
	uint8_t		dummy_bytes;
	uint8_t		data_bits;
};

static struct spi_nand_cmd_cfg alliance_cmd_cfg_table[] = {
	/*opcode                             addr_bytes addr_bits dummy_bytes data_nbits*/
	{SPINAND_CMD_GET_FEATURE,            1,         1,        0,          1},
	{SPINAND_CMD_SET_FEATURE,            1,         1,        0,          1},
	{SPINAND_CMD_PAGE_READ,              3,         1,        0,          0},
	{SPINAND_CMD_READ_FROM_CACHE,        2,         1,        1,          1},
	// {SPINAND_CMD_READ_FROM_CACHE_X2,	 2,         1,        1,          2},
	// {SPINAND_CMD_READ_FROM_CACHE_X4,	 2,         1,        1,          4},
	// {SPINAND_CMD_READ_FROM_CACHE_DUAL_IO,2, 		2,		  1,		  2},
	// {SPINAND_CMD_READ_FROM_CACHE_QUAD_IO,2,         4,        1,          4},
	{SPINAND_CMD_BLK_ERASE,              3,         1,        0,          0},
	{SPINAND_CMD_PROG_EXC,               3,         1,        0,          0},
	{SPINAND_CMD_PROG_LOAD,              2,         1,        0,          1},
	{SPINAND_CMD_PROG_LOAD_RDM_DATA,     2,         1,        0,          1},
	// {SPINAND_CMD_PROG_LOAD_X4,           2,         1,        0,          4},
	// {SPINAND_CMD_PROG_LOAD_RDM_DATA_X4,  2,         1,        0,          4},
	// {SPINAND_CMD_PROG_LOAD_RDM_QUAD,     2,         4,        0,          4},
	{SPINAND_CMD_WR_ENABLE,              0,         0,        0,          0},
	{SPINAND_CMD_WR_DISABLE,             0,         0,        0,          0},
	{SPINAND_CMD_READ_ID,                1,         1,        0,          1},
	{SPINAND_CMD_RESET,                  0,         0,        0,          0},
	{SPINAND_CMD_END},
	
};

static struct spi_nand_flash spi_nand_table[] = {
	SPI_NAND_INFO("AS5F31G04SND-08LIN",   0x52, 0x25, 2048, 64,  64, 1024, 1, 4, 0),
	SPI_NAND_INFO("AS5F32G04SND-08LIN",   0x52, 0x2E, 2048, 128, 64, 2048, 1, 8, 0),
	SPI_NAND_INFO("AS5F34G04SND-08LIN",   0x52, 0x2F, 2048, 128, 64, 4096,	1, 8, 0),
	SPI_NAND_INFO("AS5F38G04SND-08LIN",   0x52, 0x2D, 4096, 256, 64, 4096,	1, 8, 0),
	SPI_NAND_INFO("AS5F12G04SND-10LIN",   0x52, 0x8E, 2048, 128, 64, 2048,	1, 8, 0),
	SPI_NAND_INFO("AS5F14G04SND-10LIN",   0x52, 0x8F, 2048, 128, 64, 4096,	1, 8, 0),
	SPI_NAND_INFO("AS5F18G04SND-10LIN",   0x52, 0x8D, 4096, 256, 64, 4096,	1, 8, 0),
	{.name = NULL},
};

/* Private user code ---------------------------------------------------------*/


/* External variables --------------------------------------------------------*/


/* Functions -----------------------------------------------------------------*/
void FlashInit(void)
{
	printk("Starting 'Q'SPI comm with flash chip\n");

	memset(&chip, 0, sizeof(chip));

	flashSuccessfullyInitialized = false;

	spi_dev = device_get_binding("SPI_3");
	if (spi_dev == NULL) 
	{
		printk("Could not get %s device\n", "SPI_2");
		return;
	}
	else
	{
		printk("Flash SPI binding good\n");
	}

	gpioPort0 = device_get_binding("GPIO_0");

	if (gpioPort0 == NULL)
	{
		printk("Could not get gpio device\n");
		return;
	}
	else
	{
		printk("gpio0 device binding good\n");
	}

	spi_cs.gpio_dev = gpioPort0;

	// __NOP();
	// __NOP();

	int err = spi_transceive(spi_dev, &spi_cfg, &tx, &rx);

	spi_nand_init();

	flashSuccessfullyInitialized = true;
}

int spi_nand_xfer(struct spi_nand_cmd *cmd, struct spi_nand_cmd_cfg *cmd_cfg)
{
	// clear out our buffer
	//memset(m_buffer_tx,0,sizeof(m_buffer_tx));
	memset(m_buffer_rx,0,sizeof(m_buffer_rx));

	size_t totalLength = 0;
	size_t beginningOfRxData = 0;

	// set the command
	m_buffer_tx[0] = cmd->cmd;
	totalLength = 1;  // cause that's what it has to be to issue a command

	if (cmd->n_addr > 0)
	{
		totalLength = totalLength + cmd->n_addr;

		for (int x = 0; x < cmd->n_addr; x++)
		{
			m_buffer_tx[x+1] = cmd->addr[x];
		}
	}

	if (cmd_cfg->dummy_bytes > 0)
	{
		for (int x = 0; x < cmd_cfg->dummy_bytes; x++)
		{
			// add and zero out the number of dummy bytes
			m_buffer_tx[totalLength+x] = 0;
			totalLength++;
			// should only be 1 in 1 case, but in case we need to add more.
		}
	}

	if (cmd->n_tx > 0)
	{
		for (int x = 0; x < cmd->n_tx; x++)
		{
			// if there is a number of things to TX, add them
			// to our transmit buffer here 
			m_buffer_tx[totalLength+x] = cmd->tx_buf[x];
			totalLength++;
		}
	}

	if (cmd->n_rx > 0)
	{
		// mark where in the list our RX stuff should start
		beginningOfRxData = totalLength;

		for (int x = 0; x < cmd->n_rx; x++)
		{
			// if there is a number of things to RX, we just
			// need place holders to get the data
			m_buffer_tx[totalLength+x] = 0;
			totalLength++;
		}	
	}

	// this is where we tell our SPI how much stuff we tx/rx
	tx_buf.len = totalLength;
	rx_buf.len = totalLength;

	int err = spi_transceive(spi_dev, &spi_cfg, &tx, &rx);

	if (err) {
		printk("QSPI error: %d\n", err);
	} 
	else
	{
		printk("QSPI success! Command 0x%02X, length: %d\n",cmd->cmd, totalLength);

		if (cmd->n_rx > 0)
		{
			// we were looking for some data back
			for (int x = 0; x < cmd->n_rx; x++)
			{
				cmd->rx_buf[x] = m_buffer_rx[beginningOfRxData+x];
			}	
		}

	}

	return 0;
}

static struct spi_nand_cmd_cfg *spi_nand_lookup_cmd_cfg_table(uint8_t opcode,
				struct spi_nand_cmd_cfg *table)
{
	struct spi_nand_cmd_cfg *index = table;

	for (; index->opcode != SPINAND_CMD_END; index++) {
		if (index->opcode == opcode)
			return index;
	}

	printk("Invalid spi nand opcode %x\n", opcode);

	return NULL;
}

/*
 * spi_nand_issue_cmd - to process a command to send to the SPI-NAND
 * @spi: spi device structure
 * @cmd: command structure
 *
 *    Set up the command buffer to send to the SPI controller.
 *    The command buffer has to initialized to 0.
 */
int spi_nand_issue_cmd(struct spi_nand_cmd *cmd)
{
	struct spi_nand_cmd_cfg *cmd_cfg = NULL;

	cmd_cfg = spi_nand_lookup_cmd_cfg_table(cmd->cmd, alliance_cmd_cfg_table);

	if (!cmd_cfg)
	{
		return -EINVAL;
	}

	spi_nand_xfer(cmd, cmd_cfg);

	return 0;
}


/**
 * spi_nand_write_enable - send command 06h to enable write or erase the
 * Nand cells
 * @chip: SPI-NAND device structure
 * Description:
 *   Before write and erase the Nand cells, the write enable has to be set.
 *   After the write or erase, the write enable bit is automatically
 *   cleared (status register bit 2)
 *   Set the bit 2 of the status register has the same effect
 */
static int spi_nand_write_enable(void)
{
	struct spi_nand_cmd cmd;

	memset(&cmd, 0, sizeof(struct spi_nand_cmd));
	cmd.cmd = SPINAND_CMD_WR_ENABLE;

	return spi_nand_issue_cmd(&cmd);
}

/**
 * spi_nand_reset - send command FFh to reset chip.
 * @chip: SPI-NAND device structure
 */
static int spi_nand_reset(void)
{
	struct spi_nand_cmd cmd;

	memset(&cmd, 0, sizeof(struct spi_nand_cmd));
	cmd.cmd = SPINAND_CMD_RESET;

	if (spi_nand_issue_cmd(&cmd) < 0)
	{
		printk("spi_nand reset failed!\n");
	}
	else
	{
		printk("spi_nand reset success!\n");
	}

	/* elapse 2ms before issuing any other command */
	k_msleep(2);

	return 0;
}


/**
 * spi_nand_read_id - send 9Fh command to get ID
 * @chip: SPI-NAND device structure
 * @buf: buffer to store id
 */
static int spi_nand_read_id(uint8_t *buf)
{
	struct spi_nand_cmd cmd;

	memset(&cmd, 0, sizeof(struct spi_nand_cmd));
	cmd.cmd = SPINAND_CMD_READ_ID;
	cmd.n_addr = 1; 
	//cmd.addr[0] = 0; // we already set all the memory in here to 0
	cmd.n_rx = 2;
	cmd.rx_buf = buf;

	return spi_nand_issue_cmd(&cmd);
}


/**
 * spi_nand_scan_id_table - scan chip info in id table
 * @chip: SPI-NAND device structure
 * @id: point to manufacture id and device id
 * Description:
 *   If found in id table, config chip with table information.
 */
static bool spi_nand_scan_id_table(uint8_t *id)
{
	struct spi_nand_flash *type = spi_nand_table;

	for (; type->name; type++) 
	{
		if (id[0] == type->mfr_id && id[1] == type->dev_id) 
		{
			chip.name = type->name;
			chip.mfr_id = type->mfr_id;
			chip.dev_id = type->dev_id;
			chip.size = type->page_size * type->pages_per_blk * type->blks_per_lun * type->luns_per_chip;
			chip.block_size = type->page_size * type->pages_per_blk;
			chip.page_size = type->page_size;
			chip.oob_size = type->oob_size;
			//chip.lun_shift = ilog2(chip.block_size * type->blks_per_lun); // do we need?
			chip.ecc_strength = type->ecc_strength;
			chip.options = type->options;

			return true;
		}
	}

	return false;
}


/**
 * spi_nand_read_reg - send command 0Fh to read register
 * @chip: SPI-NAND device structure
 * @reg; register to read
 * @buf: buffer to store value
 */
static int spi_nand_read_reg(uint8_t reg, uint8_t *buf)
{
	struct spi_nand_cmd cmd;
	int ret;

	memset(&cmd, 0, sizeof(struct spi_nand_cmd));
	cmd.cmd = SPINAND_CMD_GET_FEATURE;
	cmd.n_addr = 1;
	cmd.addr[0] = reg;
	cmd.n_rx = 1;
	cmd.rx_buf = buf;

	ret = spi_nand_issue_cmd(&cmd);
	if (ret < 0)
		spi_nand_error("err: %d read register %d\n", ret, reg);

	return ret;
}


/**
 * spi_nand_write_reg - send command 1Fh to write register
 * @chip: SPI-NAND device structure
 * @reg; register to write
 * @buf: buffer stored value
 */
static int spi_nand_write_reg(uint8_t reg, uint8_t *buf)
{
	struct spi_nand_cmd cmd;
	int ret;

	memset(&cmd, 0, sizeof(struct spi_nand_cmd));
	cmd.cmd = SPINAND_CMD_SET_FEATURE;
	cmd.n_addr = 1;
	cmd.addr[0] = reg;
	cmd.n_tx = 1,
	cmd.tx_buf = buf,

	ret = spi_nand_issue_cmd(&cmd);
	if (ret < 0)
		spi_nand_error("err: %d write register %d\n", ret, reg);

	return ret;
}

/**
 * spi_nand_get_cfg - get configuration register value
 * @chip: SPI-NAND device structure
 * @cfg: buffer to store value
 * Description:
 *   Configuration register includes OTP config, Lock Tight enable/disable
 *   and Internal ECC enable/disable.
 */
static int spi_nand_get_cfg(uint8_t *cfg)
{
	return spi_nand_read_reg(REG_CFG, cfg);
}

/**
 * spi_nand_set_cfg - set value to configuration register
 * @chip: SPI-NAND device structure
 * @cfg: buffer stored value
 * Description:
 *   Configuration register includes OTP config, Lock Tight enable/disable
 *   and Internal ECC enable/disable.
 */
static int spi_nand_set_cfg(uint8_t *cfg)
{
	return spi_nand_write_reg(REG_CFG, cfg);
}

/**
 * spi_nand_change_mode - switch chip to OTP/OTP protect/Normal mode
 * @chip: SPI-NAND device structure
 * @mode: mode to enter
 */
static int spi_nand_change_mode(uint8_t mode)
{
	uint8_t cfg;

	spi_nand_get_cfg(&cfg);
	switch (mode) {
	case OTP_MODE:
		cfg = (cfg & ~CFG_OTP_MASK) | CFG_OTP_ENTER;
		break;
	case OTP_PROTECT_MODE:
		cfg = (cfg & ~CFG_OTP_MASK) | CFG_OTP_PROTECT;
		break;
	case SNOR_READ_ENABLE_MODE:
		cfg = (cfg & ~CFG_OTP_MASK) | CFG_SNOR_ENABLE;
		break;
	case NORMAL_MODE:
		cfg = (cfg & ~CFG_OTP_MASK) | CFG_OTP_EXIT;
		break;
	}
	spi_nand_set_cfg(&cfg);

	return 0;
}


/**
 * spi_nand_read_page_to_cache - send command 13h to read data from Nand to cache
 * @chip: SPI-NAND device structure
 * @page_addr: page to read
 */
static int spi_nand_read_page_to_cache(uint32_t page_addr)
{
	struct spi_nand_cmd cmd;

	memset(&cmd, 0, sizeof(struct spi_nand_cmd));
	cmd.cmd = SPINAND_CMD_PAGE_READ;
	cmd.n_addr = 3;
	cmd.addr[0] = (uint8_t)(page_addr >> 16);
	cmd.addr[1] = (uint8_t)(page_addr >> 8);
	cmd.addr[2] = (uint8_t)page_addr;

	return spi_nand_issue_cmd(&cmd);
}


/**
 * spi_nand_read_status - get status register value
 * @chip: SPI-NAND device structure
 * @status: buffer to store value
 * Description:
 *   After read, write, or erase, the Nand device is expected to set the
 *   busy status.
 *   This function is to allow reading the status of the command: read,
 *   write, and erase.
 *   Once the status turns to be ready, the other status bits also are
 *   valid status bits.
 */
static int spi_nand_read_status(uint8_t *status)
{
	return spi_nand_read_reg(REG_STATUS, status);
}


/**
 * spi_nand_wait - wait until the command is done
 * @chip: SPI-NAND device structure
 * @s: buffer to store status register(can be NULL)
 */
static int spi_nand_wait(uint8_t *s)
{
	unsigned long long start = k_uptime_ticks();
	uint8_t status;
	/* set timeout to 1 second */
	unsigned long long timeout = start + 1000;
	unsigned long ret = -ETIMEDOUT;

	while (k_uptime_ticks() < timeout) 
	{
		spi_nand_read_status(&status);
		if ((status & STATUS_OIP_MASK) == STATUS_READY) 
		{
			ret = 0;
			goto out;
		}
	}
out:
	if (s)
	{
		*s = status;
	}

	return ret;
}


/**
 * spi_nand_read_from_cache - read data out from cache register
 * @chip: SPI-NAND device structure
 * @page_addr: page to read
 * @column: the location to read from the cache
 * @len: number of bytes to read
 * @rbuf: buffer held @len bytes
 * Description:
 *   Command can be 03h, 0Bh, 3Bh, 6Bh, BBh, EBh
 *   The read can specify 1 to (page size + spare size) bytes of data read at
 *   the corresponding locations.
 *   No tRd delay.
 */
static int spi_nand_read_from_cache(uint32_t page_addr, uint32_t column, size_t len)
{
	struct spi_nand_cmd cmd;

	memset(&cmd, 0, sizeof(struct spi_nand_cmd));
	cmd.cmd = chip.read_cache_op;
	cmd.n_addr = 2;
	cmd.addr[0] = (uint8_t)(column >> 8);
	cmd.addr[1] = (uint8_t)column;
	cmd.n_rx = len;
	cmd.rx_buf = buffer;

	return spi_nand_issue_cmd(&cmd);
}

/**
 * spi_nand_detect_onfi - config chip with parameter page
 * @chip: SPI-NAND device structure
 * Description:
 *   This function is called when we can not get info from id table.
 */
static bool spi_nand_detect_onfi(void)
{
	struct spi_nand_onfi_params *p;
	//uint8_t buffer[256 * 3] = {0};
	int read_cache_op = 0;
	bool ret = true;
	int i;

	spi_nand_change_mode(OTP_MODE);
	spi_nand_read_page_to_cache(0x01);
	spi_nand_wait(NULL);
	/*
	* read parameter page can only ues 1-1-1 mode
	*/
	read_cache_op = chip.read_cache_op;
	chip.read_cache_op = SPINAND_CMD_READ_FROM_CACHE;

	for (int x = 509; x < (256 * 3); x++)
	{
		__NOP();
		__NOP();
		spi_nand_read_from_cache(0x01, 0, x);
	}

	__NOP();
	__NOP();
	__NOP();


	spi_nand_read_from_cache(0x01, 0, 256 * 3);
	chip.read_cache_op = read_cache_op;
	spi_nand_change_mode(NORMAL_MODE);

	p = (struct spi_nand_onfi_params *)buffer;
	for (i = 0; i < 3; i++, p++) {
		if (p->sig[0] != 'O' || p->sig[1] != 'N' ||
				p->sig[2] != 'F' || p->sig[3] != 'I')
			continue;
		if (onfi_crc16(ONFI_CRC_BASE, (uint8_t *)p, 254) ==
				le16_to_cpu(p->crc))
			break;
	}
	if (i == 3) {
		spi_nand_error("Could not find valid ONFI parameter page; aborting\n");
		ret = false;
		goto out;
	}

	memcpy(&chip.onfi_params, p, sizeof(*p));

	p = &chip.onfi_params;

	sanitize_string(p->manufacturer, sizeof(p->manufacturer));
	sanitize_string(p->model, sizeof(p->model));

	chip.name = p->model;
	chip.size = le32_to_cpu(p->byte_per_page) *
			le32_to_cpu(p->pages_per_block) *
			le32_to_cpu(p->blocks_per_lun) * p->lun_count;
	chip.block_size = le32_to_cpu(p->byte_per_page) *
			le32_to_cpu(p->pages_per_block);
	chip.page_size = le32_to_cpu(p->byte_per_page);
	chip.oob_size = le16_to_cpu(p->spare_bytes_per_page);
	//chip.lun_shift = ilog2(chip.block_size * le32_to_cpu(p->blocks_per_lun));
	chip.ecc_strength = p->vendor.alliance_sepcific.ecc_ability;

out:
	//free(buffer);
	return ret;
}


/**
 * spi_nand_init - [Interface] Init SPI-NAND device driver
 * @spi: spi device structure
 * @chip_ptr: pointer point to spi nand device structure pointer
 */
int spi_nand_init(void)
{
	uint8_t id[SPINAND_MAX_ID_LEN] = {0};

	spi_nand_reset();
	spi_nand_read_id(id);

	if (spi_nand_scan_id_table(id))
	{
		printk("success getting ID and checking with the types defined\n");
		printk(chip.name);
		printk(" found. type mfr_id: %x, dev_id: %x \n", chip.mfr_id, chip.dev_id);
	}
	else 
	{
		printk("SPI-NAND type mfr_id: %x, dev_id: %x is not in id table.\n", id[0], id[1]);
		//return -ENOTTY;  // really?  
	}
	
	spi_nand_detect_onfi();

	/*
		goto ident_done;
	spi_nand_info("SPI-NAND type mfr_id: %x, dev_id: %x is not in id table.\n",
				id[0], id[1]);

	if (spi_nand_detect_onfi(chip))
		goto ident_done;

	return -ENODEV;

ident_done:
	spi_nand_info("SPI-NAND: %s is found.\n", chip->name);

	chip->mfr_id = id[0];
	chip->dev_id = id[1];
	chip->block_shift = ilog2(chip->block_size);
	chip->page_shift = ilog2(chip->page_size);
	chip->page_mask = chip->page_size - 1;
	chip->lun = 0;

	chip->oobbuf = k_malloc(chip->oob_size);
	if (!chip->oobbuf)
		return -ENOMEM;

	if (chip->oob_size == 64)
		chip->ecclayout = &alliance_ecc_layout_64;
	else if (chip->oob_size == 128)
		chip->ecclayout = &alliance_ecc_layout_128;
	else if (chip->oob_size == 256)
		chip->ecclayout = &alliance_ecc_layout_256;

	chip->refresh_threshold = (chip->ecc_strength * 3 + 3) / 4;

	spi_nand_lock_block(chip, BL_ALL_UNLOCKED);
	spi_nand_enable_ecc(chip);
*/

	return 0;
}


/* Sanitize ONFI strings so we can safely print them */
static void sanitize_string(char *s, size_t len)
{
	int i = len - 1;
	int j = 0;

	/* Null terminate */
	s[i--] = 0;

	/* Remove unnecessary space */
	while (i >= 0 && (s[i] <= ' ' || s[i] > 127)) {
		s[i--] = 0;
	}
	/* Remove non printable chars */
	for (j = 0; j <= i; j++) {
		if (s[j] < ' ' || s[j] > 127)
			s[j] = '?';
	}
}

static uint16_t onfi_crc16(uint16_t crc, uint8_t const *p, size_t len)
{
	int i;

	while (len--) {
		crc ^= *p++ << 8;
		for (i = 0; i < 8; i++)
			crc = (crc << 1) ^ ((crc & 0x8000) ? 0x8005 : 0);
	}

	return crc;
}

/*****END OF FILE****/
